<div class="facebook-badges">
    <a class="github-button" href="https://github.com/{{theme_github_user}}/{{theme_github_repo}}" data-size="large"
        data-show-count="true" aria-label="Star {{ theme_github_user }}/{{ theme_github_repo }} on GitHub">Star</a>
</div>

<script async defer src="https://buttons.github.io/buttons.js"></script>
