<div class="meta-logo">
    <span>
        <a href="https://opensource.facebook.com"><img src="_static/opensource.png" alt="Meta Open Source" /></a>
    </span>
</div>
<div class="meta-badges">
    <a href="https://twitter.com/metaOpenSource" alt="Meta Open Source on Twitter">
        <i class="fab fa-twitter"></i>
    </a>
    &nbsp;
    <a href="https://github.com/{{theme_github_user}}/{{theme_github_repo}}" alt="View site source">
        <i class="fas fa-code"></i>
    </a>

</div>
<div class="meta-terms">
    <ul>
        <li>
            <a href="https://opensource.facebook.com/legal/terms">Terms of Use</a>
        </li>
        <li>
            <a href="https://opensource.facebook.com/legal/privacy">Privacy Policy</a>
        </li>
    </ul>
</div>

<link rel="stylesheet" href="https://use.fontawesome.com/releases/v5.15.4/css/all.css" integrity="sha384-DyZ88mC6Up2uqS4h/KRgHuoeGwBcD4Ng9SiP4dIRy0EXTlnuz47vAwmeGwVChigm" crossorigin="anonymous">
